/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.net.InetSocketAddress;
import java.nio.channels.Selector;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolver;
import com.aquenos.epics.jackie.common.io.ChannelProcessor;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.io.TimerProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.util.FutureCompletionListener;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;

/**
 * <p>
 * Threading strategy for a {@link ChannelAccessClient}. The threading strategy
 * defines how threads are created and whether components share the same thread
 * or have individual threads. It also chooses how listeners that are registered
 * by the user are notified (in the calling thread or by a separate notification
 * thread).
 * </p>
 * 
 * <p>
 * When the client creates a component, it registers it with the threading
 * strategy. The threading strategy can then decide which thread will process
 * the I/O for this component. For example, it can create a new thread, or it
 * can add the component to an existing thread.
 * </p>
 * 
 * <p>
 * Components are passed in the form of {@link CommunicationProcessor}s. The
 * threading strategy then calls the communication processor's
 * {@link CommunicationProcessor#setCommunicationController(CommunicationController)}
 * method, so that the component can register its channels and tasks that have
 * to be executed at a certain point in time. The
 * {@link CommunicationController} passed to the communication processor
 * typically is a wrapper around the {@link Selector} and timer queue for the
 * respective thread.
 * </p>
 * 
 * <p>
 * When the treading strategy detects that a channel registered by a
 * communication processor has become ready, it calls the processor's
 * {@link ChannelProcessor#processIO() processIO()} method. If a timer
 * registered by the processor has expired, it calls the processor's
 * {@link TimerProcessor#processTimer()} method. A threading strategy has to
 * assume that a communication processor is not thread safe and thus has to call
 * both methods from the same thread or ensure through other means that only one
 * thread at a time calls one of the communication processor's methods. The only
 * exception to this rule is the
 * {@link CommunicationProcessor#setCommunicationController(CommunicationController)
 * setCommunicationController(CommunicationController)} method so that the
 * threading strategy does not have to wait for the processor to return from
 * this method before it may call one of the other methods.
 * </p>
 * 
 * <p>
 * A threading strategy can choose how listeners that have been registered by
 * the user are notified: Typically, they will be called right in the thread
 * triggering the notification. However, a threading strategy might
 * alternatively choose to use a separate thread (or a pool of threads) for such
 * notifications so that a listener that blocks does not block the whole thread
 * that triggered the notification. Typically, listeners should be designed to
 * operate without blocking, but there might be applications where this property
 * cannot be guaranteed.
 * </p>
 * 
 * <p>
 * When a client is destroyed, it calls the threading strategy's
 * {@link #destroy()} method, so that the threading strategy can destroy all
 * threads. The threading strategy does not have to destroy the individual
 * communication processors as this will be done by the client.
 * </p>
 * 
 * <p>
 * Implementations of this interface must be safe for concurrent use by multiple
 * threads.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface ClientThreadingStrategy {

    /**
     * Registers a communication processor that handles a connection with a
     * specific server. Depending on the threading strategy, this might create a
     * new thread or use an existing one.
     * 
     * @param serverAddress
     *            IP address and port of the server. A threading strategy might
     *            use this information to decide which thread should be used.
     * @param communicationProcessor
     *            communication processor handling the connection with the
     *            specified server.
     */
    void addServerConnection(InetSocketAddress serverAddress,
            CommunicationProcessor communicationProcessor);

    /**
     * Unregisters a communication processor that handles a connection with a
     * specific server. A threading strategy may use this information to decide
     * when a thread it created is not used any longer and thus may be stopped.
     * There is no guarantee that the thread responsible for the specified
     * communication processor will be stopped because it might be shared with
     * other communication processors. In this case, the communication processor
     * might still be processed if one of its timers expires or one of its
     * channels becomes ready for an I/O operation. Therefore, a communication
     * processor should close all its channels and ensure that future timer
     * expirations do not result in any undesirable actions.
     * 
     * @param serverAddress
     *            IP address and port of the server. A threading strategy might
     *            use this information to decide which thread is used by the
     *            communication processor.
     * @param communicationProcessor
     *            communication processor handling the connection with the
     *            specified server.
     */
    void removeServerConnection(InetSocketAddress serverAddress,
            CommunicationProcessor communicationProcessor);

    /**
     * <p>
     * Submits a notification request for a connection listener.
     * </p>
     * 
     * <p>
     * An implementation may choose to notify the listener in the calling thread
     * (before returning) or it may choose to queue the notification request for
     * processing by a different thread. In any case, each call of this method
     * must result in exactly one corresponding call of
     * {@link ChannelAccessConnectionListener#connectionStateChanged(ChannelAccessChannel, boolean)
     * listener.connectionStateChanged(channel, nowConnected)}.
     * </p>
     * 
     * <p>
     * When an implementation chooses to notify the listener asynchronously, it
     * still has to ensure that the listener receives notifications in the
     * correct order. This means that for every call of this method using the
     * same <code>channel</code>-<code>listener</code> pair, the calls of
     * <code>listener.connectionStateChanged(...)</code> must be made in the
     * same order in which this method was called. Calls specifying different
     * listeners or channels may be reordered arbitrarily.
     * </p>
     * 
     * <p>
     * For deciding whether a channel is the same channel as specified in
     * another request and whether a listener is the same listener as specified
     * in another request, object identity should be used instead of object
     * equality. used.
     * </p>
     * 
     * @param channel
     *            channel that is the source of the event.
     * @param listener
     *            listener that shall be notified.
     * @param nowConnected
     *            new connection state of which the listener shall be notified.
     * @throws RuntimeException
     *             if the <code>listener</code> is notified synchronously and
     *             throws such an exception.
     */
    void notifyConnectionListener(ChannelAccessChannel channel,
            ChannelAccessConnectionListener listener, boolean nowConnected);

    /**
     * <p>
     * Submits an error notification request for a monitor listener.
     * </p>
     * 
     * <p>
     * An implementation may choose to notify the listener in the calling thread
     * (before returning) or it may choose to queue the notification request for
     * processing by a different thread. In any case, each call of this method
     * must result in exactly one corresponding call of
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, ChannelAccessStatus, String)
     * listener.monitorError(monitor, status, message)}.
     * </p>
     * 
     * <p>
     * When an implementation chooses to notify the listener asynchronously, it
     * still has to ensure that the listener receives notifications in the
     * correct order. This means that for every call of this method using the
     * same <code>monitor</code>-<code>listener</code> pair, the calls of
     * <code>listener.monitorError(...)</code> must be made in the same order in
     * which this method was called. Calls specifying different listeners or
     * monitors may be reordered arbitrarily.
     * </p>
     * 
     * <p>
     * For deciding whether a monitor is the same monitor as specified in
     * another request and whether a listener is the same listener as specified
     * in another request, object identity should be used instead of object
     * equality. used.
     * </p>
     * 
     * @param <T>
     *            value type of the monitor and listener.
     * @param monitor
     *            monitor that is the source of the event.
     * @param listener
     *            listener that shall be notified.
     * @param status
     *            status code that is associated with the event.
     * @param message
     *            message that is associated with the event (may be
     *            <code>null</code>).
     * @throws RuntimeException
     *             if the <code>listener</code> is notified synchronously and
     *             throws such an exception.
     */
    <T extends ChannelAccessGettableValue<?>> void notifyMonitorListenerError(
            ChannelAccessMonitor<? extends T> monitor,
            ChannelAccessMonitorListener<? super T> listener,
            ChannelAccessStatus status, String message);

    /**
     * <p>
     * Submits an event notification request for a monitor listener.
     * </p>
     * 
     * <p>
     * An implementation may choose to notify the listener in the calling thread
     * (before returning) or it may choose to queue the notification request for
     * processing by a different thread. In any case, each call of this method
     * must result in exactly one corresponding call of
     * {@link ChannelAccessMonitorListener#monitorEvent(ChannelAccessMonitor, ChannelAccessGettableValue)
     * listener.monitorEvent(monitor, value)}.
     * </p>
     * 
     * <p>
     * When an implementation chooses to notify the listener asynchronously, it
     * still has to ensure that the listener receives notifications in the
     * correct order. This means that for every call of this method using the
     * same <code>monitor</code>-<code>listener</code> pair, the calls of
     * <code>listener.monitorValue(...)</code> must be made in the same order in
     * which this method was called. Calls specifying different listeners or
     * monitors may be reordered arbitrarily.
     * </p>
     * 
     * <p>
     * For deciding whether a monitor is the same monitor as specified in
     * another request and whether a listener is the same listener as specified
     * in another request, object identity should be used instead of object
     * equality. used.
     * </p>
     * 
     * @param <T>
     *            value type of the monitor and listener. This the type of value
     *            which is passed as the event value.
     * @param monitor
     *            monitor that is the source of the event.
     * @param listener
     *            listener that shall be notified.
     * @param value
     *            event value. Typically, this is a value update for the
     *            monitor's channel.
     * @throws RuntimeException
     *             if the <code>listener</code> is notified synchronously and
     *             throws such an exception.
     */
    <T extends ChannelAccessGettableValue<?>> void notifyMonitorListenerEvent(
            ChannelAccessMonitor<? extends T> monitor,
            ChannelAccessMonitorListener<? super T> listener, T value);

    /**
     * <p>
     * Submits a notification request for a future completion listener.
     * </p>
     * 
     * <p>
     * An implementation may choose to notify the listener in the calling thread
     * (before returning) or it may choose to queue the notification request for
     * processing by a different thread. In any case, each call of this method
     * must result in exactly one corresponding call of
     * {@link FutureCompletionListener#completed(ListenableFuture)
     * listener.completed(future)}.
     * </p>
     * 
     * <p>
     * An implementation may choose to reorder notification requests
     * arbitrarily.
     * </p>
     * 
     * @param <T>
     *            type of the future's value.
     * @param future
     *            future that completed and is the source of the event.
     * @param listener
     *            listener that shall be notified.
     */
    <T> void notifyFutureCompletionListener(ListenableFuture<T> future,
            FutureCompletionListener<? super T> listener);

    /**
     * Sets the communication processor for the {@link BeaconDetector}. As there
     * is only one beacon detector for each client, the threading strategy does
     * not have to maintain a list, but can simply assume that this method is
     * only called once.
     * 
     * @param communicationProcessor
     *            communication processor used by the beacon detector.
     */
    void setBeaconDetector(CommunicationProcessor communicationProcessor);

    /**
     * Sets the communication processor used for the {@link ChannelNameResolver}
     * . As there is only one channel name resolver for each client, the
     * threading strategy does not have to maintain a list, but can simply
     * assume that this method is only called once.
     * 
     * @param communicationProcessor
     *            communication processor used by the channel name resolver.
     */
    void setChannelNameResolver(CommunicationProcessor communicationProcessor);

    /**
     * Destroys the threading strategy. This means that the threading strategy
     * should stop all threads and release all resources acquired internally. It
     * does not have to destroy the individual communication processors. This
     * method should not block waiting for the shutdown of all threads to
     * complete, but return immediately once the shutdown of the threads has
     * been initiated.
     */
    void destroy();

}
