/*
 * Copyright 2015-2019 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.io.IOException;
import java.net.InetSocketAddress;

import com.aquenos.epics.jackie.client.ChannelAccessClientConfiguration;
import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.io.ChannelAccessServerConnection;
import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.io.TimerProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessAccessRightsMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessCancelSubscriptionClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessCancelSubscriptionServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessChannelDisconnectedByServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConnectChannelClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConnectChannelFailedMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConnectChannelServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessDisconnectChannelMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessErrorMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventsOffMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventsOnMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessReadNotifyClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessReadNotifyServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSubscriptionClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSubscriptionServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessWriteMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessWriteNotifyClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessWriteNotifyServerMessage;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Connection to a Channel Access TCP server that acts as a PV server. This
 * class is only intended for use by the classes inside this package and by the
 * {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * This class is closely coupled to the {@link Server}: Each server has exactly
 * one connection and each connection has exactly one server. The functionality
 * is separated into two classes for three reasons: First, both classes are
 * already quite large, thus uniting them in a single class would create an
 * extremely large class. Second, the connection class has to inherit from
 * {@link ChannelAccessServerConnection} which would be inconvenient for the
 * {@link Server}. Third, the two objects have different scopes. This connection
 * handles everything that is related to I/O. In particular, most operations run
 * inside the communication thread and it has to be created inside the
 * communication thread (because it creates socket channels from its
 * constructor). This design allows the code to work without having to acquire a
 * mutex. The server, on the other hand, is created outside the communication
 * thread (the communication thread might not even exist yet when the server is
 * created). For this reasons, the server has a mutex to protect access to its
 * fields.
 * </p>
 * 
 * <p>
 * Most of the logic implemented inside this class runs inside the communication
 * thread. It will also dispatch events to the {@link Server} within this
 * thread. The only methods that may run outside the communication thread are
 * the {@link #destroy()} method and the methods that send messages. Those
 * methods are thread safe even though they are lock free.
 * </p>
 * 
 * <p>
 * A connection is destroyed when its channel is closed (e.g. because the
 * connection to the server is interrupted) or when its {@link Server} is
 * destroyed.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessPVServerConnection
        extends ChannelAccessServerConnection {

    /**
     * Connection state.
     * 
     * @author Sebastian Marsching
     */
    public enum State {

        /**
         * Connection has been created but not fully established yet. This is
         * the initial state.
         */
        CONNECTION_PENDING,

        /**
         * Connection has been established and is responsive.
         */
        CONNECTION_RESPONSIVE,

        /**
         * Connection has been established, but is unresponsive.
         */
        CONNECTION_UNRESPONSIVE,

        /**
         * Connection has been destroyed. This means that the channel has been
         * closed.
         */
        CONNECTION_DESTROYED;

    }

    private static final int MAX_MESSAGE_SIZE = Integer.MAX_VALUE
            - (Integer.MAX_VALUE
                    % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);

    private CommunicationController communicationController;
    private ChannelAccessClientConfiguration configuration;
    private boolean gracefulDestructionRequested;
    private Server server;
    private InetSocketAddress serverAddress;
    private State state = State.CONNECTION_PENDING;
    private boolean subscriptionEventsDisabled = false;
    private ClientThreadingStrategy threadingStrategy;

    /**
     * Creates a connection to a Channel Access PV server.
     * 
     * @param beaconDetector
     *            beacon detector to be used by the watch dog that checks this
     *            connection's responsiveness.
     * @param communicationController
     *            that controls the communication thread. The
     *            {@link CommunicationProcessor CommunicationProcessors} that
     *            are created for this connection are registered with this
     *            controller.
     * @param configuration
     *            client configuration. This is used to get the configuration
     *            options (e.g. the character set) needed by the connection.
     * @param server
     *            server instance for which this connection is created.
     * @param serverAddress
     *            socket address of the Channel Access PV server.
     * @param threadingStrategy
     *            threading strategy that is responsible for this connection.
     *            The threading strategy is notified when this connection is
     *            destroyed, so that the corresponding thread can be stopped.
     * @param version
     *            version of the Channel Access protocol that should initially
     *            be used for communication with the server. This version is
     *            typically determined from the response to the search requests
     *            which led to this server connection to be created. While
     *            protocol upgrades are supported, protocol downgrades are not.
     *            This means that the connection will fail if the server
     *            responds with an older Channel Access version.
     * @throws IOException
     *             if the socket channel for the connection cannot be created.
     */
    public ChannelAccessPVServerConnection(BeaconDetector beaconDetector,
            CommunicationController communicationController,
            ChannelAccessClientConfiguration configuration, Server server,
            InetSocketAddress serverAddress,
            ClientThreadingStrategy threadingStrategy,
            ChannelAccessVersion version) throws IOException {
        super(beaconDetector, configuration.getCharset(),
                communicationController,
                Math.round(configuration.getEchoInterval() * 1000.0),
                configuration.getErrorHandler(),
                calculateMaxMessageSize(
                        configuration.getMaxPayloadReceiveSize()),
                calculateMaxMessageSize(configuration.getMaxPayloadSendSize()),
                serverAddress, configuration.getHostName(),
                configuration.getUserName(), version);
        this.communicationController = communicationController;
        this.configuration = configuration;
        this.server = server;
        this.serverAddress = serverAddress;
        this.threadingStrategy = threadingStrategy;
    }

    private static int calculateMaxMessageSize(int maxPayloadSize) {
        if (maxPayloadSize < 65536) {
            return maxPayloadSize + ChannelAccessConstants.STANDARD_HEADER_SIZE;
        } else {
            if (maxPayloadSize < MAX_MESSAGE_SIZE
                    - ChannelAccessConstants.LARGE_HEADER_SIZE) {
                return maxPayloadSize
                        + ChannelAccessConstants.LARGE_HEADER_SIZE;
            } else {
                return MAX_MESSAGE_SIZE;
            }
        }
    }

    private void destroyGracefullyInternal() {
        // In contrast to the destroyGracefully() method, this method is always
        // called within the communication thread.
        if (isSendQueueEmpty()) {
            // If the send queue is already empty, we can (and should) destroy
            // the connection immediately.
            destroy();
        } else {
            // If there is still data in the send queue, we want to send it
            // first. The connection is going to be destroyed when all data has
            // been sent. We have checked that there is data left, so we can be
            // sure that onSend() is going to be called.
            gracefulDestructionRequested = true;
        }
    }

    private void updateState(State newState) {
        // We never go back from the destroyed state to a different state.
        if (this.state.equals(State.CONNECTION_DESTROYED)) {
            return;
        }
        if (!this.state.equals(newState)) {
            this.state = newState;
            this.server.updateConnectionState(newState);
        }
    }

    @Override
    public boolean isDestroyed() {
        // We only override this method to make it public.
        return super.isDestroyed();
    }

    @Override
    protected void onConnect() {
        // We have to call the method from the super class before doing anything
        // else. This ensures that the watchdog is started and the version
        // message is sent.
        super.onConnect();
        if (isDestroyed()) {
            // If the connection has already been destroyed, we do not want to
            // do anything.
            return;
        }
        // Now we are ready for operation. We have to mark the connection as
        // responsive so that other commands (in particular CA_PROTO_CREATE_CHAN
        // commands) can be sent.
        updateState(State.CONNECTION_RESPONSIVE);
    }

    @Override
    protected void onConnectionResponsive() {
        // This is equivalent to responsiveCircuitNotify in the C library.
        updateState(State.CONNECTION_RESPONSIVE);
        server.connectionResponsive();
    }

    @Override
    protected void onConnectionUnresponsive() {
        // This is equivalent to unresponsiveCircuitNotify in the C library.
        updateState(State.CONNECTION_UNRESPONSIVE);
        server.connectionUnresponsive();
    }

    @Override
    protected void onDestroy() {
        // Clean up data structures in super class.
        super.onDestroy();
        // We have to unregister the connection from the threading strategy so
        // that the corresponding thread may be stopped.
        threadingStrategy.removeServerConnection(serverAddress, server);
        // This method is called in the thread that called destroy(). Usually,
        // this is the communication thread, unless the server's destroy()
        // method has been called. The server's destroy() method gets called
        // when the connect attempt fails (in the communication thread), the
        // server pool is destroyed (in this case the communication thread has
        // already been stopped), or when there are no channels left with the
        // server. The first case is fine because our assumption that only the
        // communication thread updates the state holds. The second case is fine
        // because there is no other thread that could potentially update the
        // state. Finally, the third case is also fine because the server sets
        // its internal state to destroyed before destroying this connection.
        // This means that any other state update is ignored by the server and
        // we can be sure that the server's state will not be messed up by the
        // communication thread. This does not hold for the connection's state,
        // however we do not really care about the connection's state if the
        // server has been properly destroyed, because the connection will not
        // be used any longer anyway.
        updateState(State.CONNECTION_DESTROYED);
    }

    @Override
    protected void onMessageReceived(ChannelAccessMessage message) {
        // If this connection is going to be destroyed, we ignore any incoming
        // message. Our corresponding Server object has already been destroyed
        // and thus there is no sense in forwarding the message.
        if (gracefulDestructionRequested) {
            return;
        }
        switch (message.getCommand()) {
        case CA_PROTO_VERSION:
            // The version message is already handled by the super class: If the
            // version sent by the server is less than the expected version
            // (that we know from the name lookup), the connection is destroyed.
            break;
        case CA_PROTO_EVENT_ADD:
            // This message type is actually used for two kinds of messages, so
            // we have to check which kind of message we received.
            if (message instanceof ChannelAccessSubscriptionServerMessage) {
                processSubscriptionMessage(
                        (ChannelAccessSubscriptionServerMessage) message);
            } else {
                processCancelSubscriptionMessage(
                        (ChannelAccessCancelSubscriptionServerMessage) message);
            }
            break;
        case CA_PROTO_ERROR:
            processErrorMessage((ChannelAccessErrorMessage) message);
            break;
        case CA_PROTO_CLEAR_CHANNEL:
            // Most servers do not send this message. We already clear the
            // resources when sending the disconnect message to the server, so
            // we can simply ignore this message.
            break;
        case CA_PROTO_READ_NOTIFY:
            processReadNotifyMessage(
                    (ChannelAccessReadNotifyServerMessage) message);
            break;
        case CA_PROTO_CREATE_CHAN:
            processConnectChannelMessage(
                    (ChannelAccessConnectChannelServerMessage) message);
            break;
        case CA_PROTO_WRITE_NOTIFY:
            processWriteNotifyMessage(
                    (ChannelAccessWriteNotifyServerMessage) message);
            break;
        case CA_PROTO_ACCESS_RIGHTS:
            processAccessRightsMessage(
                    (ChannelAccessAccessRightsMessage) message);
            break;
        case CA_PROTO_ECHO:
            // This message is already handled by the super class, so we can
            // ignore it.
            break;
        case CA_PROTO_CREATE_CH_FAIL:
            processConnectChannelFailedMessage(
                    (ChannelAccessConnectChannelFailedMessage) message);
            break;
        case CA_PROTO_SERVER_DISCONN:
            processChannelDisconnectedByServerMessage(
                    (ChannelAccessChannelDisconnectedByServerMessage) message);
            break;
        case CA_PROTO_READ:
        case CA_PROTO_READ_SYNC:
            // We never send these messages to the server, so we do not expect
            // them in a response.
        default:
            // If an unexpected message is received, we destroy the connection.
            // We do this, because we cannot know what consequences this message
            // might have.
            configuration.getErrorHandler().handleError(
                    ChannelAccessPVServerConnection.class, null,
                    "Received unexpected message from TCP server: " + message);
            destroy();
        }
    }

    @Override
    protected void onReceiveFinished(boolean moreDataAvailable) {
        // First we call the implementation from the super class. This is
        // important for the echo watchdog.
        super.onReceiveFinished(moreDataAvailable);
        // For a connection with a PV server, we have to use flow control when
        // too many subscription events are received.
        if (moreDataAvailable && !subscriptionEventsDisabled) {
            // We are receiving too much data, enable flow control.
            sendMessage(new ChannelAccessEventsOffMessage());
            subscriptionEventsDisabled = true;
        } else if (!moreDataAvailable && subscriptionEventsDisabled) {
            // The data flow is okay again, disable flow control.
            sendMessage(new ChannelAccessEventsOnMessage());
            subscriptionEventsDisabled = false;
        }
        // If we disabled subscription events, we have to make sure that
        // tryReceive() is called again. It is possible, that we just read the
        // last byte of data and in this case no more data would (because events
        // are disabled) and this method would not get called (and thus we would
        // not enable subscription events). We schedule a timer that calls
        // tryReceive(). This timer will not be processed before the current I/O
        // processing finishes. This way, trySend() also has a chance to do its
        // work.
        if (subscriptionEventsDisabled) {
            communicationController.registerTimer(new TimerProcessor() {
                @Override
                public void processTimer() {
                    tryReceive();
                }
            }, 0L);
        }
    }

    @Override
    protected void onSend() {
        // Most actions are handled by the super class. We only override this
        // method in order to destroy this connection when graceful destruction
        // has been requested and there is no more data to be sent.
        super.onSend();
        if (gracefulDestructionRequested) {
            if (isSendQueueEmpty()) {
                destroy();
            }
        }
    }

    private void processSubscriptionMessage(
            ChannelAccessSubscriptionServerMessage message) {
        server.subscriptionEventReceived(message.getSubscriptionId(),
                message.getStatus(), message.getValue());
    }

    private void processCancelSubscriptionMessage(
            ChannelAccessCancelSubscriptionServerMessage message) {
        server.subscriptionCancelled(message.getSubscriptionId());
    }

    private void processErrorMessage(ChannelAccessErrorMessage message) {
        ChannelAccessMessage originalMessage = message.getOriginalMessage();
        if (originalMessage == null) {
            // If we do not have the original message, we cannot act on the
            // error, so we log and ignore it.
            configuration.getErrorHandler().handleError(this.getClass(),
                    new ChannelAccessException(message.getStatus(),
                            message.getMessageString()),
                    "Received an error message that did not have a originating message attached to to it.");
            return;
        }
        switch (originalMessage.getCommand()) {
        case CA_PROTO_EVENT_ADD:
            server.subscriptionErrorReceived(
                    ((ChannelAccessSubscriptionClientMessage) originalMessage)
                            .getSubscriptionId(),
                    message.getStatus(), message.getMessageString());
            break;
        case CA_PROTO_READ_NOTIFY:
            server.readErrorReceived(
                    ((ChannelAccessReadNotifyClientMessage) originalMessage)
                            .getOperationId(),
                    message.getStatus(), message.getMessageString());
            break;
        case CA_PROTO_WRITE:
            // Unlike for a CA_PROTO_WRITE_NOTIFY message, the operation that
            // sent the CA_PROTO_WRITE message completed instantaneously, so
            // there is no direct way of reporting this error to the calling
            // code. However, the server instance might have additional
            // information (e.g. the name of the channel, if the channel
            // instance still exists), so we pass the message on to it.
            server.writeNoNotifyErrorReceived(
                    ((ChannelAccessWriteMessage) originalMessage)
                            .getChannelCID(),
                    message.getStatus(), message.getMessageString());
            break;
        case CA_PROTO_WRITE_NOTIFY:
            server.writeErrorReceived(
                    ((ChannelAccessWriteNotifyClientMessage) originalMessage)
                            .getOperationId(),
                    message.getStatus(), message.getMessageString());
            break;
        // The C library also handles CA_PROTO_READ and CA_PROTO_WRITE, but we
        // do not have to handle these because we never send them.
        default:
            // The server should not send an error message in response to
            // another message. If it does, we log the error and ignore it.
            // The C library does the same.
            configuration.getErrorHandler().handleError(this.getClass(),
                    new ChannelAccessException(message.getStatus(),
                            message.getMessageString()),
                    "Received an unexpected error message in response to a  "
                            + originalMessage.getCommand() + " message.");
        }
    }

    private void processReadNotifyMessage(
            ChannelAccessReadNotifyServerMessage message) {
        server.readResponseReceived(message.getOperationId(),
                message.getStatus(), message.getValue());

    }

    private void processConnectChannelMessage(
            ChannelAccessConnectChannelServerMessage message) {
        server.channelConnected(message.getChannelCID(),
                message.getChannelSID(), message.getNativeDataType(),
                message.getNativeCount());
    }

    private void processWriteNotifyMessage(
            ChannelAccessWriteNotifyServerMessage message) {
        server.writeResponseReceived(message.getOperationId(),
                message.getStatus());
    }

    private void processAccessRightsMessage(
            ChannelAccessAccessRightsMessage message) {
        server.receivedAccessRights(message.getChannelCID(),
                message.isMayRead(), message.isMayWrite());

    }

    private void processConnectChannelFailedMessage(
            ChannelAccessConnectChannelFailedMessage message) {
        server.connectChannelFailed(message.getChannelCID());
    }

    private void processChannelDisconnectedByServerMessage(
            ChannelAccessChannelDisconnectedByServerMessage message) {
        server.channelDisconnectedByServer(message.getChannelCID());
    }

    /**
     * Sends a {@link ChannelAccessConnectChannelClientMessage}. This method is
     * only intended for use by the {@link Server} associated with this
     * connection. However, it may be called from any thread and is lock free.
     * 
     * @param cid
     *            client-assigned identifier for the channel that shall be
     *            connected.
     * @param channelName
     *            network name of the channel.
     */
    void connectChannel(int cid, String channelName) {
        sendMessage(new ChannelAccessConnectChannelClientMessage(cid,
                ChannelAccessVersion.NEWEST_SUPPORTED_VERSION, channelName));
    }

    /**
     * Sends a {@link ChannelAccessDisconnectChannelMessage}. This method is
     * only intended for use by the {@link Server} associated with this
     * connection. However, it may be called from any thread and is lock free.
     * 
     * @param cid
     *            client-assigned identifier for the channel that shall be
     *            disconnected.
     * @param sid
     *            server-assigned identifier for the channel that shall be
     *            disconnected. This is the identifier that the server sent when
     *            connecting the channel.
     */
    void disconnectChannel(int cid, int sid) {
        sendMessage(new ChannelAccessDisconnectChannelMessage(cid, sid));
    }

    /**
     * Sends a {@link ChannelAccessReadNotifyClientMessage}. This method is only
     * intended for use by the {@link Server} associated with this connection.
     * However, it may be called from any thread and is lock free.
     * 
     * @param sid
     *            server-assigned identifier of the channel which shall be read.
     *            This is the identifier that the server sent when connecting
     *            the channel.
     * @param dataType
     *            data type that shall be used in the response from the server.
     * @param count
     *            number of elements that shall be sent by the server. May only
     *            be zero for Channel Access version 4.13 or newer. In this case
     *            it means the channel's native count.
     * @param operationId
     *            client-assigned identifier for the operation. This is used by
     *            the client (or in our case the {@link Server} instance) to map
     *            a response from the server to the request.
     * @throws ChannelAccessException
     *             if the read attempt would result in a response that exceeds
     *             the maximum payload size for received messages.
     */
    void readChannel(int sid, ChannelAccessValueType dataType, int count,
            int operationId) {
        // Check whether the reply to the read request would result in a
        // response that is too large.
        verifyResponsePayloadSize(dataType, count);
        sendMessage(new ChannelAccessReadNotifyClientMessage(dataType, count,
                sid, operationId));
    }

    /**
     * Sends a {@link ChannelAccessWriteNotifyClientMessage}. This method is
     * only intended for use by the {@link Server} associated with this
     * connection. However, it may be called from any thread and is lock free.
     * 
     * @param sid
     *            server-assigned identifier of the channel which shall be
     *            written. This is the identifier that the server sent when
     *            connecting the channel.
     * @param value
     *            value that shall be sent to the server.
     * @param operationId
     *            client-assigned identifier for the operation. This is used by
     *            the client (or in our case the {@link Server} instance) to map
     *            a response from the server to the request.
     * @throws ChannelAccessException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     */
    void writeChannel(int sid, ChannelAccessPuttableValue<?> value,
            int operationId) {
        try {
            sendMessage(
                    new ChannelAccessWriteNotifyClientMessage(value.getType(),
                            value.getValueSize(), sid, operationId, value));
        } catch (SerializedPayloadTooLargeException e) {
            throw new ChannelAccessException(
                    ChannelAccessStatus.StatusMessage.ECA_TOLARGE.toStatus());
        }
    }

    /**
     * Sends a {@link ChannelAccessWriteMessage}. This method is only intended
     * for use by the {@link Server} associated with this connection. However,
     * it may be called from any thread and is lock free.
     * 
     * @param cid
     *            client-assigned identifier for the channel that shall be
     *            written.
     * @param sid
     *            server-assigned identifier of the channel which shall be
     *            written. This is the identifier that the server sent when
     *            connecting the channel.
     * @param value
     *            value that shall be sent to the server.
     * @throws ChannelAccessException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     */
    void writeChannelNoNotify(int cid, int sid,
            ChannelAccessPuttableValue<?> value) {
        try {
            sendMessage(new ChannelAccessWriteMessage(value.getType(),
                    value.getValueSize(), sid, cid, value));
        } catch (SerializedPayloadTooLargeException e) {
            throw new ChannelAccessException(
                    ChannelAccessStatus.StatusMessage.ECA_TOLARGE.toStatus());
        }
    }

    /**
     * Sends a {@link ChannelAccessSubscriptionClientMessage}. This method is
     * only intended for use by the {@link Server} associated with this
     * connection. However, it may be called from any thread and is lock free.
     * 
     * @param sid
     *            server-assigned identifier of the channel for which the
     *            subscription shall be requested. This is the identifier that
     *            the server sent when connecting the channel.
     * @param dataType
     *            data type that shall be used in responses from the server.
     * @param count
     *            number of elements that shall be sent by the server. May only
     *            be zero for Channel Access version 4.13 or newer. In this case
     *            it means the channel's native count.
     * @param mask
     *            event mask that shall be used for the subscription. This
     *            defines which events the server is going to send.
     * @param subscriptionId
     *            client-assigned identifier for the subscription. This is used
     *            by the client (or in our case the {@link Server} instance) to
     *            map a response from the server to the corresponding
     *            subscription.
     * @throws ChannelAccessException
     *             if the event notifications would result in responses that
     *             exceed the maximum payload size for received messages.
     */
    void subscribe(int sid, ChannelAccessValueType dataType, int count,
            ChannelAccessEventMask mask, int subscriptionId) {
        // Check whether the event notifications would result in responses that
        // are too large.
        verifyResponsePayloadSize(dataType, count);
        sendMessage(new ChannelAccessSubscriptionClientMessage(dataType, count,
                sid, subscriptionId, mask));
    }

    /**
     * Sends a {@link ChannelAccessCancelSubscriptionClientMessage}. This method
     * is only intended for use by the {@link Server} associated with this
     * connection. However, it may be called from any thread and is lock free.
     * 
     * @param sid
     *            server-assigned identifier of the channel for which we want to
     *            cancel a subscription. This is the identifier that the server
     *            sent when connecting the channel.
     * @param dataType
     *            data type that was used when requesting the subscription.
     * @param count
     *            element count that was used when requesting the subscription.
     * @param subscriptionId
     *            client-assigned identifier for the subscription. This must be
     *            the identifier that was used when requesting the subscription.
     */
    void unsubscribe(int sid, ChannelAccessValueType dataType, int count,
            int subscriptionId) {
        sendMessage(new ChannelAccessCancelSubscriptionClientMessage(dataType,
                count, sid, subscriptionId));
    }

    @Override
    public void destroy() {
        // We only override this method to make it public.
        super.destroy();
    }

    public void destroyGracefully() {
        // If the connection has already been destroyed, there is nothing we
        // have to do.
        if (isDestroyed()) {
            return;
        }
        // We mark this connection as shutting down and destroy it when there is
        // no more data to be sent. If we are not in the communication thread,
        // we register a timer, because some of the actions (including setting
        // the destruction flag) can only be run in the communication thread.
        if (communicationController.inCommunicationThread()) {
            destroyGracefullyInternal();
        } else {
            communicationController.registerTimer(new TimerProcessor() {
                @Override
                public void processTimer() {
                    destroyGracefullyInternal();
                }
            }, 0L);
        }
        // If the connection does not shutdown gracefully within 30 seconds
        // (e.g. because the server does not accept the data we send and thus
        // the queue cannot be flushed), we kill the connection the hard way.
        communicationController.registerTimer(new TimerProcessor() {
            @Override
            public void processTimer() {
                destroy();
            }
        }, System.currentTimeMillis() + 30000L);
    }

}
