/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.ShortPayloadException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Representation of a <code>CA_PROTO_EVENT_CANCEL</code> message that is sent
 * from a client to a server. This message is sent in order to cancel an event
 * subscription.
 * 
 * @see ChannelAccessCommand#CA_PROTO_EVENT_CANCEL
 * @see ChannelAccessCancelSubscriptionServerMessage
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessCancelSubscriptionClientMessage extends
        ChannelAccessMessage {

    private short dataType;
    private int count;
    private int sid;
    private int subscriptionId;

    /**
     * Creates a <code>CA_PROTO_EVENT_CANCEL</code> message. This constructor
     * should be used to create a message that is supposed to be sent from a
     * client to a server.
     * 
     * @param dataType
     *            data-type that was specified when creating the subscription.
     * @param count
     *            element count that was specified when creating the
     *            subscription.
     * @param sid
     *            server-assigned ID identifying the channel.
     * @param subscriptionId
     *            client-assigned ID identifying the subscription. This must be
     *            the ID that was specified when creating the subscription.
     */
    public ChannelAccessCancelSubscriptionClientMessage(
            ChannelAccessValueType dataType, int count, int sid,
            int subscriptionId) {
        super(ChannelAccessCommand.CA_PROTO_EVENT_CANCEL);

        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.sid = sid;
        this.subscriptionId = subscriptionId;
    }

    private ChannelAccessCancelSubscriptionClientMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, int channelSID, int subscriptionId) {
        super(ChannelAccessCommand.CA_PROTO_EVENT_CANCEL, messageHeader);

        this.dataType = dataType;
        this.count = count;
        this.sid = channelSID;
        this.subscriptionId = subscriptionId;
    }

    /**
     * Returns the data-type for the subscription. This should be the same data
     * type that was specified when the subscription was originally requested.
     * 
     * @return data-type for the subscription or <code>null</code> if the
     *         numeric identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            return ChannelAccessValueType.forTypeNumber(dataType);
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of elements of the value that should be sent in
     * subscription events. This should be the same number that was specified
     * when the subscription was originally requested.
     * 
     * @return number of value elements for the subscription..
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the channel SID. This is the identifier that the server uses to
     * identify the channel. This number might be negative or zero.
     * 
     * @return channel SID assigned to the channel by the server.
     */
    public int getChannelSID() {
        return sid;
    }

    /**
     * Returns the client-assigned subscription ID. The subscription ID is used
     * by the client in order to map events sent in response to a subscription
     * to the respective subscription. This number might be negative or zero.
     * 
     * @return client-assigned subscription identifier.
     */
    public int getSubscriptionId() {
        return subscriptionId;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, dataType, count, sid, subscriptionId,
                version, maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        verifyHeader(0, count, version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessCancelSubscriptionClientMessage other = (ChannelAccessCancelSubscriptionClientMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.sid, other.sid)
                .append(this.subscriptionId, other.subscriptionId).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(sid)
                .append(subscriptionId).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", sid=");
        sb.append(sid);
        sb.append(", subscriptionId=");
        sb.append(subscriptionId);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws ShortPayloadException
     *             if the message's payload size is less than 16 bytes and
     *             <code>headerOnly</code> is <code>false</code>.
     */
    protected static ChannelAccessCancelSubscriptionClientMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_EVENT_CANCEL
                .getCommandNumber());
        if (!headerOnly) {
            byteSource.skip(messageHeader.getPayloadSize());
        }
        return new ChannelAccessCancelSubscriptionClientMessage(messageHeader,
                messageHeader.getDataType(), messageHeader.getCount(),
                messageHeader.getCID(), messageHeader.getContextSpecific());
    }

}
