/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_CREATE_CH_FAIL</code> message. This
 * message is sent from a server to a client via TCP if the request to connect a
 * channel could not be fulfilled.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_CREATE_CH_FAIL
 */
public final class ChannelAccessConnectChannelFailedMessage extends
        ChannelAccessMessage {

    private int cid;

    /**
     * Constructor for creating a <code>CA_PROTO_CREATE_CH_FAIL</code> message.
     * 
     * @param cid
     *            the channel CID. This is the identifier used by the client to
     *            identify the channel.
     */
    public ChannelAccessConnectChannelFailedMessage(int cid) {
        super(ChannelAccessCommand.CA_PROTO_CREATE_CH_FAIL);

        this.cid = cid;
    }

    private ChannelAccessConnectChannelFailedMessage(
            ChannelAccessMessageHeader messageHeader, int cid) {
        super(ChannelAccessCommand.CA_PROTO_CREATE_CH_FAIL, messageHeader);

        this.cid = cid;
    }

    /**
     * Returns the channel CID used by the client. This is the identifier that
     * the client uses to identify the channel. This number might be negative or
     * zero.
     * 
     * @return channel CID assigned to the channel by the client.
     */
    public int getChannelCID() {
        return cid;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, (short) 0, 0, cid, 0, version,
                maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessConnectChannelFailedMessage other = (ChannelAccessConnectChannelFailedMessage) obj;
        return this.cid == other.cid;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including the payload).
     */
    protected static ChannelAccessConnectChannelFailedMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_CREATE_CH_FAIL
                .getCommandNumber());
        byteSource.skip(messageHeader.getPayloadSize());
        return new ChannelAccessConnectChannelFailedMessage(messageHeader,
                messageHeader.getCID());
    }

}
