/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.apache.commons.lang3.StringUtils;

/**
 * Utility methods for dealing with IPv4 addresses.
 * 
 * @author Sebastian Marsching
 */
public final class Inet4AddressUtil {

    private Inet4AddressUtil() {
    }

    /**
     * The loopback address (127.0.0.1).
     */
    public static final Inet4Address LOOPBACK_ADDRESS;
    static {
        try {
            LOOPBACK_ADDRESS = (Inet4Address) InetAddress
                    .getByAddress(new byte[] { 127, 0, 0, 1 });
        } catch (UnknownHostException e) {
            throw new RuntimeException(
                    "Unexpected exception when trying to generate loopback address.");
        }
    }

    /**
     * Converts an IPv4 address from a series of four bytes to an integer. The
     * bytes must be in network byte-order (big endian).
     * 
     * @param address
     *            IPv4 address as a series of four bytes.
     * @return IPv4 address as an integer.
     */
    public static int addressToInt(byte[] address) {
        if (address.length != 4) {
            throw new IllegalArgumentException(
                    "An IPv4 address must have 4 bytes, but this address has "
                            + address.length + " bytes.");
        }
        int intAddress = address[3] & 0xFF;
        intAddress |= ((address[2] << 8) & 0xFF00);
        intAddress |= ((address[1] << 16) & 0xFF0000);
        intAddress |= ((address[0] << 24) & 0xFF000000);
        return intAddress;
    }

    /**
     * Converts from an IPv4 address from an integer to a series of four bytes.
     * The bytes will be in network byte-order (big endian).
     * 
     * @param address
     *            IPv4 address as an integer.
     * @return IPv4 address as a series of four bytes.
     */
    public static byte[] addressToBytes(int address) {
        byte[] byteAddress = new byte[4];
        byteAddress[0] = (byte) ((address >>> 24) & 0xFF);
        byteAddress[1] = (byte) ((address >>> 16) & 0xFF);
        byteAddress[2] = (byte) ((address >>> 8) & 0xFF);
        byteAddress[3] = (byte) (address & 0xFF);
        return byteAddress;
    }

    /**
     * Parses a string representing an IPv4 address. The address must be a valid
     * IPv4 address. In particular, it must only contain digits and the period
     * (.), must consist of exactly four parts divided by periods, and each part
     * must represent a number between 0 and 255. Leading zeros in parts are
     * allowed, however there may not be more than three digits per part. Each
     * part must have at least one digit.
     * 
     * @param addressString
     *            IPv4 address in typical notation (NNN.NNN.NNN.NNN).
     * @return IPv4 address created from string.
     * @throws IllegalArgumentException
     *             if <code>addressString</code> is not a valid IPv4 address.
     */
    public static Inet4Address stringToInetAddress(String addressString) {
        String[] addressParts = StringUtils.split(addressString, '.');
        if (addressParts.length != 4) {
            throw new IllegalArgumentException("Not a valid IPv4 address: "
                    + addressString);
        }
        byte[] addressBytes = new byte[4];
        for (int i = 0; i < addressBytes.length; i++) {
            String addressPart = addressParts[i];
            if (addressPart.length() < 1 || addressPart.length() > 3
                    || !StringUtils.isNumeric(addressPart)) {
                throw new IllegalArgumentException("Not a valid IPv4 address: "
                        + addressString);
            }
            int numericPart = Integer.parseInt(addressPart, 10);
            // We do not have to check for negative values because we know that
            // the string only contains digits.
            if (numericPart > 255) {
                throw new IllegalArgumentException("Not a valid IPv4 address: "
                        + addressString);
            }
            addressBytes[i] = (byte) numericPart;
        }
        try {
            return (Inet4Address) InetAddress.getByAddress(addressBytes);
        } catch (UnknownHostException e) {
            // This should never happen, because we verified the address string.
            throw new RuntimeException(e);
        }
    }

}
