/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;

/**
 * <p>
 * Policy applied when dealing with listener objects.
 * </p>
 * 
 * <p>
 * When dealing with listeners that can be removed (unregistered) there is an
 * inherent conflict of interest: A listener might be called to be notified of
 * an event while another thread tries to remove this exact listener. In general
 * there are three different ways how one can deal with this situation:
 * </p>
 * 
 * <ol>
 * <li>The thread trying to remove the listener is blocked until the
 * notification has completed. This has the effect that the listener is never
 * called after the remove operation has succeeded: Either it is removed before
 * being called and thus not called, or the call is completed before the remove
 * method returns. This policy is represented by the {@link #BLOCK} value. While
 * this policy is the easiest to understand, as it provides strong consistency
 * guarantees, it is prone to potential deadlocks (see below).</li>
 * <li>The thread trying to remove the listener can simply remove it and return.
 * However, this means that a listener might be called once after being removed.
 * This happens when the listener is already in the process of being called,
 * when it is removed because the thread calling the listener might not be aware
 * that it has just been removed. This policy is represented by the
 * {@link #IGNORE} value.</li>
 * <li>The thread trying to remove the listener can remove it, but signals the
 * problem to the caller by throwing a {@link ConcurrentNotificationException}.
 * This policy is represented by the {@link #REPORT} value. Like in case of the
 * {@link #IGNORE} policy, the listener is still removed.</li>
 * </ol>
 * 
 * <h3>Risk of deadlocks when using the {@code BLOCK} policy</h3>
 * 
 * <p>
 * When using the {@link #BLOCK} policy, the code calling the listener has to
 * hold a lock (lock A) while calling the listener. The code removing the
 * listener has to acquire the same lock. Fundamentally, this is not dangerous,
 * but danger might arise if the listener also acquires a lock (lock B). Now, if
 * the code removing the listener holds the same lock (lock B) while trying to
 * remove the listener, there is the risk of a deadlock. This happens when the
 * code trying to remove the listener acquires the external lock (lock B) while
 * another thread acquires the internal lock (lock A) in order to notify the
 * listener. Now the thread trying to remove the listener cannot continue
 * because it has to acquire lock A in order to do so, but this lock is held by
 * the thread calling the listener. The thread calling the listener cannot
 * continue either, because the listener tries to acquire lock B, which is
 * already held by the thread trying to remove the listener.
 * </p>
 * 
 * <p>
 * This means that when using the {@code BLOCK} policy, it is very important to
 * not acquire a lock in a listener, that might also be held by code calling the
 * remove method. If such a situation is unavoidable or not in the sphere of
 * influence of the developer (e.g. because a listener has to be removed in
 * response to a request made by some external code potentially holding a lock
 * and some external code potentially acquire the same lock has to be called
 * from a listener), the {@link #IGNORE} or {@link #REPORT} policy should be
 * used instead.
 * </p>
 * 
 * @since EPICS Jackie 2.0.0
 * 
 * @author Sebastian Marsching
 */
public enum ListenerLockPolicy {

    /**
     * Wait for an internal lock when removing a listener. This ensures that the
     * listener is never called after the remove method returns, but might be
     * susceptible to deadlocks. Please refer to the
     * {@linkplain ListenerLockPolicy class description} for details.
     */
    BLOCK,

    /**
     * Do not wait for the internal lock when removing a listener. If the lock
     * cannot be acquired, the listener is still removed, but it might be called
     * at most once after the remove method has returned. This policy is similar
     * to the {@link #REPORT} policy, but unlike that policy does not notify the
     * caller of the fact that the listener might still be called one more time.
     */
    IGNORE,

    /**
     * Do not wait for the internal lock when removing a listener, but report
     * the problem. If the lock cannot be acquired, the listener is still
     * removed, but it might be called at most once after the remove method has
     * returned. The remove method throws an
     * {@link ConcurrentNotificationException} in order to notify the calling
     * code of this situation. This policy is similar to the {@link #IGNORE}
     * policy, but unlike this policy, the {@code IGNORE} policy does not cause
     * an exception and thus the calling code is not notified.
     */
    REPORT

}
